@@ -51,7 +51,8 @@ class Serializer(object):
         self.writesock = writesock
 
     def send(self, msg):
-        buf = msgpack.packb(msg, use_bin_type=True)
+        buf = msgpack.packb(msg, use_bin_type=True,
+                            unicode_errors='surrogateescape')
         self.writesock.sendall(buf)
 
     def close(self):
@@ -64,7 +65,9 @@ class Serializer(object):
 class Deserializer(six.Iterator):
     def __init__(self, readsock):
         self.readsock = readsock
-        self.unpacker = msgpack.Unpacker(use_list=False, encoding='utf-8')
+        self.unpacker = msgpack.Unpacker(use_list=False, raw=False,
+                                         strict_map_key=False,
+                                         unicode_errors='surrogateescape')
 
     def __iter__(self):
         return self
@@ -74,10 +77,13 @@ class Deserializer(six.Iterator):
             try:
                 return next(self.unpacker)
             except StopIteration:
-                buf = self.readsock.recv(4096)
-                if not buf:
-                    raise
-                self.unpacker.feed(buf)
+                try:
+                    buf = self.readsock.recv(4096)
+                    if not buf:
+                        raise
+                    self.unpacker.feed(buf)
+                except socket.timeout:
+                    pass
 
 
 class Future(object):
@@ -108,6 +114,7 @@ class Future(object):
 
 class ClientChannel(object):
     def __init__(self, sock):
+        self.running = False
         self.writer = Serializer(sock)
         self.lock = threading.Lock()
         self.reader_thread = threading.Thread(
@@ -122,13 +129,17 @@ class ClientChannel(object):
 
     def _reader_main(self, reader):
         """This thread owns and demuxes the read channel"""
+        with self.lock:
+            self.running = True
         for msg in reader:
             msgid, data = msg
             if msgid is None:
                 self.out_of_band(data)
             else:
                 with self.lock:
-                    assert msgid in self.outstanding_msgs
+                    if msgid not in self.outstanding_msgs:
+                        raise AssertionError("msgid should in "
+                                             "outstanding_msgs.")
                     self.outstanding_msgs[msgid].set_result(data)
 
         # EOF.  Perhaps the privileged process exited?
@@ -141,6 +152,7 @@ class ClientChannel(object):
         with self.lock:
             for mbox in self.outstanding_msgs.values():
                 mbox.set_exception(exc)
+            self.running = False
 
     def out_of_band(self, msg):
         """Received OOB message. Subclasses might want to override this."""
@@ -151,7 +163,8 @@ class ClientChannel(object):
         future = Future(self.lock)
 
         with self.lock:
-            assert myid not in self.outstanding_msgs
+            if myid in self.outstanding_msgs:
+                raise AssertionError("myid shoudn't be in outstanding_msgs.")
             self.outstanding_msgs[myid] = future
             try:
                 self.writer.send((myid, msg))
