@@ -12,7 +12,7 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
-'''Privilege separation ("privsep") daemon.
+"""Privilege separation ("privsep") daemon.
 
 To ease transition this supports 2 alternative methods of starting the
 daemon, all resulting in a helper process running with elevated
@@ -41,8 +41,9 @@ privileges and open socket(s) to the original process:
 The privsep daemon exits when the communication channel is closed,
 (which usually occurs when the unprivileged process exits).
 
-'''
+"""
 
+from concurrent import futures
 import enum
 import errno
 import io
@@ -55,24 +56,45 @@ import sys
 import tempfile
 import threading
 
-if platform.system() == 'Linux':
-    import fcntl
-    import grp
-    import pwd
-
 import eventlet
+from eventlet import patcher
 from oslo_config import cfg
 from oslo_log import log as logging
+from oslo_utils import encodeutils
 from oslo_utils import importutils
+import six
 
 from oslo_privsep._i18n import _
 from oslo_privsep import capabilities
 from oslo_privsep import comm
 
+if platform.system() == 'Linux':
+    import fcntl
+    import grp
+    import pwd
 
 LOG = logging.getLogger(__name__)
 
 
+EVENTLET_MODULES = ('os', 'select', 'socket', 'thread', 'time', 'MySQLdb',
+                    'builtins', 'subprocess')
+EVENTLET_LIBRARIES = []
+
+
+def _null():
+    return []
+
+
+for module in EVENTLET_MODULES:
+    if hasattr(patcher, '_green_%s_modules' % module):
+        method = getattr(patcher, '_green_%s_modules' % module)
+    elif hasattr(patcher, '_green_%s' % module):
+        method = getattr(patcher, '_green_%s' % module)
+    else:
+        method = _null()
+    EVENTLET_LIBRARIES.append((module, method))
+
+
 @enum.unique
 class StdioFd(enum.IntEnum):
     # NOTE(gus): We can't use sys.std*.fileno() here.  sys.std*
@@ -206,7 +228,9 @@ class _ClientChannel(comm.ClientChannel):
     def out_of_band(self, msg):
         if msg[0] == Message.LOG:
             # (LOG, LogRecord __dict__)
-            record = pylogging.makeLogRecord(msg[1])
+            message = {encodeutils.safe_decode(k): v
+                       for k, v in msg[1].items()}
+            record = pylogging.makeLogRecord(message)
             if LOG.isEnabledFor(record.levelno):
                 LOG.logger.handle(record)
         else:
@@ -254,6 +278,21 @@ def replace_logging(handler, log_root=None):
     log_root.addHandler(handler)
 
 
+def un_monkey_patch():
+    for eventlet_mod_name, func_modules in EVENTLET_LIBRARIES:
+        if not eventlet.patcher.is_monkey_patched(eventlet_mod_name):
+            continue
+
+        for name, mod in func_modules():
+            patched_mod = sys.modules.get(name)
+            orig_mod = eventlet.patcher.original(name)
+            for attr_name in mod.__patched__:
+                patched_attr = getattr(mod, attr_name, None)
+                unpatched_attr = getattr(orig_mod, attr_name, None)
+                if patched_attr is not None:
+                    setattr(patched_mod, attr_name, unpatched_attr)
+
+
 class ForkingClientChannel(_ClientChannel):
     def __init__(self, context):
         """Start privsep daemon using fork()
@@ -275,6 +314,7 @@ class ForkingClientChannel(_ClientChannel):
 
         if os.fork() == 0:
             # child
+            un_monkey_patch()
 
             channel = comm.ServerChannel(sock_b)
             sock_a.close()
@@ -352,6 +392,9 @@ class Daemon(object):
         self.user = context.conf.user
         self.group = context.conf.group
         self.caps = set(context.conf.capabilities)
+        self.thread_pool = futures.ThreadPoolExecutor(
+            context.conf.thread_pool_size)
+        self.communication_error = None
 
     def run(self):
         """Run request loop. Sets up environment, then calls loop()"""
@@ -413,22 +456,75 @@ class Daemon(object):
                 'inh': fmt_caps(inh),
             })
 
-    def _process_cmd(self, cmd, *args):
+    def _process_cmd(self, msgid, cmd, *args):
+        """Executes the requested command in an execution thread.
+
+        This executes a call within a thread executor and returns the results
+        of the execution.
+
+        :param msgid: The message identifier.
+        :param cmd: The `Message` type indicating the command type.
+        :param args: The function, args, and kwargs if a Message.CALL type.
+        :return: A tuple of the return status, optional call output, and
+                 optional error information.
+        """
         if cmd == Message.PING:
             return (Message.PONG.value,)
 
-        elif cmd == Message.CALL:
+        try:
+            if cmd != Message.CALL:
+                raise ProtocolError(_('Unknown privsep cmd: %s') % cmd)
+
+            # Extract the callable and arguments
             name, f_args, f_kwargs = args
             func = importutils.import_class(name)
-
             if not self.context.is_entrypoint(func):
                 msg = _('Invalid privsep function: %s not exported') % name
                 raise NameError(msg)
 
             ret = func(*f_args, **f_kwargs)
             return (Message.RET.value, ret)
+        except Exception as e:
+            LOG.debug(
+                'privsep: Exception during request[%(msgid)s]: '
+                '%(err)s', {'msgid': msgid, 'err': e}, exc_info=True)
+            cls = e.__class__
+            cls_name = '%s.%s' % (cls.__module__, cls.__name__)
+            return (Message.ERR.value, cls_name, e.args)
 
-        raise ProtocolError(_('Unknown privsep cmd: %s') % cmd)
+    def _create_done_callback(self, msgid):
+        """Creates a future callback to receive command execution results.
+
+        :param msgid: The message identifier.
+        :return: A future reply callback.
+        """
+        channel = self.channel
+
+        def _call_back(result):
+            """Future execution callback.
+
+            :param result: The `future` execution and its results.
+            """
+            try:
+                reply = result.result()
+                LOG.debug('privsep: reply[%(msgid)s]: %(reply)s',
+                          {'msgid': msgid, 'reply': reply})
+                channel.send((msgid, reply))
+            except IOError:
+                self.communication_error = sys.exc_info()
+            except Exception as e:
+                LOG.debug(
+                    'privsep: Exception during request[%(msgid)s]: '
+                    '%(err)s', {'msgid': msgid, 'err': e}, exc_info=True)
+                cls = e.__class__
+                cls_name = '%s.%s' % (cls.__module__, cls.__name__)
+                reply = (Message.ERR.value, cls_name, e.args)
+                try:
+                    channel.send((msgid, reply))
+                except IOError:
+                    self.communication_error = sys.exc_info()
+
+        return _call_back
 
     def loop(self):
         """Main body of daemon request loop"""
@@ -439,27 +535,16 @@ class Daemon(object):
         self.context.set_client_mode(False)
 
         for msgid, msg in self.channel:
-            LOG.debug('privsep: request[%(msgid)s]: %(req)s',
-                      {'msgid': msgid, 'req': msg})
-            try:
-                reply = self._process_cmd(*msg)
-            except Exception as e:
-                LOG.debug(
-                    'privsep: Exception during request[%(msgid)s]: %(err)s',
-                    {'msgid': msgid, 'err': e}, exc_info=True)
-                cls = e.__class__
-                cls_name = '%s.%s' % (cls.__module__, cls.__name__)
-                reply = (Message.ERR.value, cls_name, e.args)
-
-            try:
-                LOG.debug('privsep: reply[%(msgid)s]: %(reply)s',
-                          {'msgid': msgid, 'reply': reply})
-                self.channel.send((msgid, reply))
-            except IOError as e:
-                if e.errno == errno.EPIPE:
+            error = self.communication_error
+            if error:
+                if error[1].errno == errno.EPIPE:
                     # Write stream closed, exit loop
                     break
-                raise
+                six.reraise(*error)
+
+            # Submit the command for execution
+            future = self.thread_pool.submit(self._process_cmd, msgid, *msg)
+            future.add_done_callback(self._create_done_callback(msgid))
 
         LOG.debug('Socket closed, shutting down privsep daemon')
 
